local function create_render_target(name, w, h)
    local color_params = {
        format = render.FORMAT_RGBA,
        width = w,
        height = h,
        min_filter = render.FILTER_NEAREST,
        mag_filter = render.FILTER_NEAREST,
        u_wrap = render.WRAP_CLAMP_TO_BORDER,
        v_wrap = render.WRAP_CLAMP_TO_BORDER
    }

    local depth_params = {
        format = render.FORMAT_DEPTH,
        width = w,
        height = h,
        min_filter = render.FILTER_NEAREST,
        mag_filter = render.FILTER_NEAREST,
        u_wrap = render.WRAP_CLAMP_TO_BORDER,
        v_wrap = render.WRAP_CLAMP_TO_BORDER
    }

    return render.render_target(name, { [render.BUFFER_COLOR_BIT] = color_params, [render.BUFFER_DEPTH_BIT] = depth_params })
end

local function update(self)
    local pov = math.atan(render.get_height() * 0.5 / 1000) * 2
    local aspect_ratio = render.get_width() / render.get_height()

    local near_width  = self.near_scale * pov
    local near_height = near_width / aspect_ratio
    local far_width   = self.far_scale * pov
    local far_height  = far_width / aspect_ratio

    local forward_vector = vmath.normalize(vmath.rotate(self.camera.rotation, vmath.vector3(0, 0, -1)))
    local up_vector = vmath.rotate(self.camera.rotation, vmath.vector3(0, 1, 0))
    local right_vector = vmath.cross(forward_vector, up_vector)

    local near_center = forward_vector * self.near_scale --+ self.camera.position
    local far_center  = forward_vector * self.far_scale  --+ self.camera.position

    local points = {
        near_top_left     = near_center +  up_vector * near_height + -right_vector * near_width,
        near_top_right    = near_center +  up_vector * near_height +  right_vector * near_width,
        near_bottom_left  = near_center + -up_vector * near_height + -right_vector * near_width,
        near_bottom_right = near_center + -up_vector * near_height +  right_vector * near_width,
        far_top_left      = far_center  +  up_vector * far_height  + -right_vector * far_width,
        far_top_right     = far_center  +  up_vector * far_height  +  right_vector * far_width,
        far_bottom_left   = far_center  + -up_vector * far_height  + -right_vector * far_width,
        far_bottom_right  = far_center  + -up_vector * far_height  +  right_vector * far_width,
    }

    local centroid = vmath.vector3()
    local distance = -math.huge

    for _, point in pairs(points) do
        centroid = centroid + point / 8
    end
    for _, point in pairs(points) do
        distance = math.max(vmath.length(point - centroid), distance)
    end

    centroid = centroid + self.camera.position

    self.view_matrix = vmath.matrix4_look_at(centroid + -self.light.direction * distance, centroid, vmath.vector3(0, 1, 0))

    local min_x, min_y, min_z =  math.huge,  math.huge,  math.huge
    local max_x, max_y, max_z = -math.huge, -math.huge, -math.huge

    for i, point in pairs(points) do
        local pos = self.view_matrix * vmath.vector4(point.x, point.y, point.z, 0)

        min_x = math.min(min_x, pos.x)
        max_x = math.max(max_x, pos.x)
        min_y = math.min(min_y, pos.y)
        max_y = math.max(max_y, pos.y)
        min_z = math.min(min_z, pos.z)
        max_z = math.max(max_z, pos.z)
    end

    self.projection_matrix = vmath.matrix4_orthographic(min_x/2, max_x/2, min_y/2, max_y/2, 0, (max_z - min_z) * 2)
end

local function get_render_target(self)
    return self.render_target
end

local function get_projection_matrix(self)
    return self.projection_matrix
end

local function get_view_matrix(self)
    return self.view_matrix
end

local M = {}

M.new = function(camera, light)
    local state = {
        camera = camera,
        light = light,

        projection_matrix = vmath.matrix4(),
        view_matrix = vmath.matrix4(),
        near_scale = 0.1,
        far_scale = 35,
        render_target = create_render_target("shadowmap", 1024 * 2, 1024 * 2),

        update = update,

        get_render_target = get_render_target,
        get_projection_matrix = get_projection_matrix,
        get_view_matrix = get_view_matrix,
    }

    return state
end

return M
